#include "madcalc.h"
#include "newmadcalc.h"
#include "ui_madcalc.h"
#include "global.h"
#include "qmath.h" //Библиотека для поддержки математических операций, таких как sqrt и pow
#include "QDialog"
#include "button.h"
#include "prog.h"
#include "QDesktopWidget"
#include "QFile"
#include "QTextStream"
#include "QDebug"
#include <QDir>
#include <QProcess>
#include <stdio.h>
#include <curl/curl.h> //Данная библиотека обеспечивает скачивание файлов по разным протоколам, в том числе HTTPS
#include <string>
#include <QThread>
#include <QMessageBox>

size_t write_data(void *ptr, size_t size, size_t nmemb, FILE *stream) {
    size_t written = fwrite(ptr, size, nmemb, stream);
    return written;
}


madCalc::madCalc(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::madCalc)
{
    setWindowFlags(Qt::Dialog | Qt::MSWindowsFixedSizeDialogHint); //Данная строчка запрещает изменение размеров окна

    ui->setupUi(this);
    //Сразу конструируем форму справки, чтобы потом ее можно было просто показывать или прятать одной строчкой
    //Если поместить этот конструктор в какую-нибудь отдельную функцию, то потом она будет открываться по кнопке вызова
    //Столько раз сколько вы ее нажмете. А так раз нажмете - и вот она, висит и все. Повторное нажатие уже не вызовет
    //ее дубликат
    formAbout = new about;
    mUpdate   = new madUpdate;

    socket = new QTcpSocket; //Это для подключения к серверу автообновления

    //Передаем версию программы, полученную с сервера обновления, в форму madUpdate
    connect(this, &madCalc::signalVersion, mUpdate, &madUpdate::versionResieve);
    connect(mUpdate, &madUpdate::updateState, this, &madCalc::updateState); //Данное выражение получает сигнал из второй формы

    //TCP connect signal slot
    QObject::connect(socket, &QTcpSocket::readyRead, this, &madCalc::socket_Read_Data);
    QObject::connect(socket, &QTcpSocket::disconnected, this, &madCalc::socket_Disconnected);

    //connect(ui->sendBtn, SIGNAL(clicked()), this, SLOT(sendSTR()));

    //Далее подключаем цифровые кнопки
    connect(ui->btn0, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn1, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn2, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn3, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn4, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn5, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn6, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn7, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn8, SIGNAL(clicked()), this, SLOT(digitClicked()));
    connect(ui->btn9, SIGNAL(clicked()), this, SLOT(digitClicked()));

    //Кнопки унарных операций
    connect(ui->btn1x, SIGNAL(clicked()), this,  SLOT(unaryOperatorClicked()));  //Кнопка btnSin вычисляет 1/x

    //Вычисление корня
    connect(ui->btnRoot, SIGNAL(clicked()), this, SLOT(root()));  //Кнопка для вычисления корня

    //Операции сложения (+ и -)
    connect(ui->btnPlus, SIGNAL(clicked()), this, SLOT(Operator()));
    connect(ui->btnMinus, SIGNAL(clicked()), this, SLOT(Operator()));

    //Мультипликативные операции
    connect(ui->btnDivide, SIGNAL(clicked()), this, SLOT(Operator()));
    connect(ui->btnMultiply, SIGNAL(clicked()), this, SLOT(Operator()));

    //Кнопка +/-
    connect(ui->btnPM, SIGNAL(clicked()), this, SLOT(funcPM()));

    //Кнопки сброса
    connect(ui->btnC, SIGNAL(clicked()), this, SLOT(ClearAll())); //Кнопка С
    connect(ui->btnCE, SIGNAL(clicked()), this, SLOT(CE()));//Кнопка CE

    //Кнопка "." для цифрового ввода
    connect(ui->btnDot, SIGNAL(clicked()), this, SLOT(btnDot()));

    //Кнопка которая удалет последний знак в строке lcd1
    connect(ui->btnBackspace, SIGNAL(clicked()), this, SLOT(Backspase()));

    //Кнопка =
    connect(ui->btnEqual, SIGNAL(clicked()), this, SLOT(equalClicked()));

    //Кнопки управления памятью
    connect(ui->btnMC, SIGNAL(clicked()), this, SLOT(clearMemory()));
    connect(ui->btnMR, SIGNAL(clicked()), this, SLOT(readMemory()));
    connect(ui->btnMP, SIGNAL(clicked()), this, SLOT(addToMemory()));
    connect(ui->btnMM, SIGNAL(clicked()), this, SLOT(subtMemory()));
    connect(ui->btnMS, SIGNAL(clicked()), this, SLOT(setMemory()));

    //Кнопка вычисления процентов
    connect(ui->btnPer, SIGNAL(clicked()), this,  SLOT(Percent()));

    //Кнопка вызова формы About
    connect(ui->btnAb, SIGNAL(clicked()), this,  SLOT(About()));

    //Кнопка вызова второй формы
    connect(ui->btnEn, SIGNAL(clicked()), this, SLOT(En()));      //Кнопка вызова инженерного режима
    //Кнопка вызова второй формы
    connect(ui->btnProg, SIGNAL(clicked()), this, SLOT(Prog()));  //Кнопка вызова режима программиста

    //Здесь довольно сложная конструкция статус бара. Он состоит из двух QLabel,
    //которые перекрывают друг друга, в каждом свое выравнивание
    ui->status1->setStyleSheet(
    "QLabel{background-color: rgba(255, 255, 255, 0);}"
    "qproperty-alignment: 'AlignVCenter | AlignLeft';"
    );
    ui->status2->setStyleSheet(
    "QLabel{background-color: rgba(255, 255, 255, 0);"
    "qproperty-alignment: 'AlignVCenter | AlignRight';"
    "color: black;"
    "border: 1px solid gray;"
    "}"
    );
    ui->status1->setText(" madmentat.ru");
    ui->status2->setText("Для справки нажмите F1");
    ui->lcd3->setText("");
    waitingForOperand = true;
    moveToCenter();
    updateCheck(); //Инициируем проверку доступности обновления
  //update();
}

void madCalc::moveToCenter(){
    QRect rect = frameGeometry();
    rect.moveCenter(QDesktopWidget().availableGeometry().center());
    move(rect.topLeft());
}

//Начало блока обновления программы
void madCalc::updateState(QString uState) //Здесь мы принимаем решение пользователя об обновлении
{
    if (uState == "Yes"){ //Если пользователь согласился, принимаем Yes
    upState = uState;
    qDebug() << "User has decided to update!";
    update();             //И затем инициируем функцию update()
    }
}

void madCalc::update() //Здесь мы вызовем вспомогательную программу-обновлялку и закроем madCalc
{

#ifdef Q_OS_LINUX //задаем условия компиляции для Linux
        QString PATH = QCoreApplication::applicationDirPath(); //Получим полный путь к madCalc и запишем в переменную PATH
        qDebug() << "Linux-based operation system determinated";
        qDebug() << "Working path is: " + PATH;
        QProcess process;
        process.setProgram("update");
        process.setWorkingDirectory(PATH);
        process.setStandardOutputFile(QProcess::nullDevice());
        process.setStandardErrorFile(QProcess::nullDevice());
        qint64 pid;
        process.startDetached(&pid);
        QApplication::quit();
#endif

}

void madCalc::updateCheck()
{
#ifdef Q_OS_LINUX //задаем условия компиляции для Linux
        QString myname = "madCalc Linux";
#endif
#ifdef Q_OS_WIN //задаем условия компиляции для Windows
        QString myname = "madCalc Win";
#endif

        QString IP = "madmentat.ru";
        int port = 5555;
        socket->abort();
        //Connecting servers
        socket->connectToHost(IP, port);
        //Waiting for the connection to succeed
        if(!socket->waitForConnected(30000))
        {
            qDebug() << "Connection failed!";
            return;
        }
        qDebug() << "Connect successfully!";
        qDebug() << "Send request to https://update.madmentat.ru: " << myname;

        //Берем имя программы из myname и отсылаем на сервер as ASCII code
        socket->write(myname.toLatin1());
        socket->flush();
        socket_Read_Data();
}

void madCalc::socket_Read_Data()
{
    QByteArray buffer;
    buffer = socket->readAll();   //Read Buffer Data
    if(!buffer.isEmpty())
    {
         QString str = buffer;
         qDebug() << "Current madCalc version received: " << buffer;
         //Disconnect
         socket->disconnectFromHost();
     }

     if (buffer > version){
         mUpdate->show();
         emit signalVersion(buffer);
     }
}



void madCalc::socket_Disconnected()
{
    //Здесь практически нефига не делается, только добавим в дебак сообщение
    qDebug() << "Disconnected!";
}

//Окончание блока обновление программы

void madCalc::closeEvent(QCloseEvent *event)
{
    formAbout->close();
}

void madCalc::En()
{
    this->hide();
    newmadcalc *frm = new newmadcalc();
                           frm->show();
                    formAbout->close();
                    //При переходе в другой режим калькулятора закрываем форму справки,
                    //чтобы повторное нажатие F1 не вызвало дубликат формы справки
}


void madCalc::Prog()
{
    this->hide();
    prog *frm = new prog();
    frm->show();
}

madCalc::~madCalc()
{
    delete ui;
    delete this->socket;
}

void madCalc::StatusBar()
{
    //Выводим в статус бар сообщение о содержимом памяти, но чтобы это получилось, для начала переводим double в qstring
    QString valueAsString = QString::number(sumInMemory);
    ui->status1->setText(" Число в памяти: " + valueAsString);
}


//Кнопка backspase
void madCalc::Backspase()
{
    double operand = ui->lcd1->text().toDouble(); //Переписываем содержимое экрана в переменную operand типа double
    if (operand == 0){                            //Если операнд на экране равен нулю, то
        ui->lcd1->setText("0");                   //Оставляем все как есть
    }
    else {                                        //В противном случае подрезаем его справа
    QString str;
    str = ui->lcd1->text();
    str.resize(str.size() - 1);                   //На единицу
    ui->lcd1->setText(str);                       //И возвращаем результат на экран
    }
    int sizeOfText = ui->lcd1->text().size();     //Здесь мы объявляем переменную sizeOfText которая принимает кол-во символов в строке
    if (sizeOfText < 1) {                         //И запускаем проверку. Если на экране остался меньше 1 символа, то
        ui->lcd1->setText("0");                   //Записываем на экран 0
    }
    waitingForOperand = true;                     //А вот без этой строчки при дальнейшем наборе циферек будет нечто типа 0123 и т. п.
}

//Блок управления памятью
//MC стирает данные из ячейки памяти.
void madCalc::clearMemory() {
 sumInMemory = 0.0;

    ui->status1->setText(" madmentat.ru");
    ui->status2->setText("Для справки нажмите F1");

    if (ui->lcd1->text() == "0"){waitingForOperand = true;} //Данная проверка нужна для того чтобы после нажатия кнопки избежать нуля перед числом, вроде 0123
   else waitingForOperand = false;
}
//Читает из памяти и выводит на экран
void madCalc::readMemory() {
 ui->lcd1->setText(QString::number(sumInMemory));
 if (ui->lcd1->text() == "0"){waitingForOperand = true;} //Данная проверка нужна для того чтобы после нажатия кнопки избежать нуля перед числом, вроде 0123
else waitingForOperand = false;
}
void madCalc::setMemory() {
 sumInMemory = ui->lcd1->text().toDouble();
 StatusBar();
 if (ui->lcd1->text() == "0"){waitingForOperand = true;} //Данная проверка нужна для того чтобы после нажатия кнопки избежать нуля перед числом, вроде 0123
else waitingForOperand = false;
 }

//M+ прибавляет к числу из памяти число, отображенное на дисплее и записывает результат в память вместо предыдущего
void madCalc::addToMemory() {
 sumInMemory += ui->lcd1->text().toDouble();
 StatusBar();
 if (ui->lcd1->text() == "0"){waitingForOperand = true;} //Данная проверка нужна для того чтобы после нажатия кнопки избежать нуля перед числом, вроде 0123
else waitingForOperand = false;
}
//MS - сохраняет в память число, отображенное на дисплее
void madCalc::subtMemory() {
 sumInMemory -= ui->lcd1->text().toDouble();
 StatusBar();
 if (ui->lcd1->text() == "0"){waitingForOperand = true;} //Данная проверка нужна для того чтобы после нажатия кнопки избежать нуля перед числом, вроде 0123
else waitingForOperand = false;
 }

//Функция кнопки +/-
void madCalc::funcPM()
{
    QPushButton *button = (QPushButton *)sender();
        double all_numbers;
        QString new_label;
        if(button->text() == "+/-"){                            //Если текст кнопки содержит "+/-", то выполняем следующее:
            all_numbers = (ui->lcd1->text().toDouble());        //Переводим содержимое lcd1 в формат Double и записываем в переменную all_numbers
            all_numbers = all_numbers * -1;                     //Умножаем all_numbers на -1
            new_label = QString::number(all_numbers, 'g', 13);  //Ограничиваем строку в 13 знаков
            ui->lcd1->setText(new_label);                       // выводим сообщение в lcd1
        }
}

//Сообщение об ошибке в случае ошибки (например, деление на 0)
void madCalc::abortOperation() {
    //Обнуляем все наши переменные
    num = 0.0;
   // factorSoFar = 0.0;
    pendingAdditiveOperator.clear();
    pendingMultiplicativeOperator.clear();
    //Ждем следующего операнда
    waitingForOperand = true;
    //Выводим на экран сообщение об ошибке
 ui->lcd1->setText(tr("ERROR"));
}

//Ввод с клавиатуры
void madCalc::keyPressEvent(QKeyEvent *event) {
 int key=event->key();                   //event->key() - целочисленный код клавиши
 QString str = QString(QChar(key));
 if (key>=Qt::Key_0 && key<=Qt::Key_9) { //Цифровые клавиши 0..9
 digitValue = str.toDouble();
 Digit();
 }
 else if (key==Qt::Key_Backspace) { //BackSpace стирает символ
  Backspase();
 }
 else if (key==Qt::Key_Delete) { //Delete стирает всё
  ClearAll();
 }

 else if (key==Qt::Key_Plus || key==Qt::Key_Minus || key==Qt::Key_Asterisk || key==Qt::Key_Slash ) {
 clickedOperator = str;
 OperatorSender();
 }
 else if (key==Qt::Key_Enter) {
 equalClicked();
 }

 else if (key==Qt::Key_Period) {
btnDot();
 }

 else if (key==Qt::Key_F1) {
 About(); //Функция вызова формы about
 }
}

//Конец конструкции ввода с клавиатуры

//Данная функция принимает сигналы от кнопок из GUI, а затем записывает их содержание в digitValue и вызывает метод Digit()
void madCalc::digitClicked() {
 QPushButton *button = (QPushButton *)sender();
 digitValue = button->text().toInt();
 Digit();
}

//Метод Digit()
void madCalc::Digit(){
 if (ui->lcd1->text() == "0" && digitValue == 0.0) return;
 if (waitingForOperand) {
  ui->lcd1->clear();
  waitingForOperand = false;
 }
 ui->lcd1->setText(ui->lcd1->text() + QString::number(digitValue));
 ui->lcd3->setText("");
}

//Актуальная конструкция для точки Dot(".")
void madCalc::btnDot() {
 if (waitingForOperand) ui->lcd1->setText("0");
 if (!ui->lcd1->text().contains(".")) ui->lcd1->setText(ui->lcd1->text() + tr("."));
 waitingForOperand = false;
}

//Вычисление процента
void madCalc::Percent()
{
 double operand = ui->lcd1->text().toDouble(); // берем операнд c экрана
 PER = num*operand/100;  // Вычисляем процент num от operand и сохраняем это в переменную PER
 ui->lcd1->setText(QString::number(PER));      // Выводим результат на экран
}

//Вычисление корня
void madCalc::root()
{
    QPushButton *button = (QPushButton *)sender(); //Слушаем sender
    clickedOperator = button->text();
    if (clickedOperator == tr("²√x")){
    double operand = ui->lcd1->text().toDouble();  //Записываем в operand содержимое lcd1
    num = operand;
    double result = 0.0;                           //объявляем переменную result, равную 0
    if (operand == -1){
        QString complex = "i";
        ui->lcd1->setText(complex); //Выводим result на экран
    }
    else if (operand < -1) {
         operand = -1 * operand;
          result = sqrt(operand);
          QString str = QString::number(result);
          QString complex = str + "i";
          ui->lcd1->setText(complex); //Выводим result на экран
     }
    else {
         result = sqrt(operand);
         QString str = QString::number(result);
         ui->lcd1->setText(str);     //Выводим result на экран
     }
}
    QString str = QString::number(num);
    ui->lcd2->setText("√(" + str + ") =");
}

//Унарные операции
void madCalc::unaryOperatorClicked() {
    ui->lcd3->setText("");

 QPushButton *button = (QPushButton *)sender(); //Слушаем sender
 clickedOperator = button->text();              //Пишем в clickedOperator знак, прочитанный с кнопки
 double operand = ui->lcd1->text().toDouble();  //Записываем в operand содержимое lcd1
 double result = 0.0;                           //объявляем переменную result, равную 0
 if (clickedOperator == tr("√")) { //Запускаем проверку: если корень извлекается из отрицательного числа, запускаем abortOperation();
  if (operand < 0.0) { // Заострить внимание на этом моменте, в дальнейшем я бы хотел запилить поддержку комплексных чисел!
   abortOperation();
   return;
  }
  result = sqrt(operand); //А вообще вычисляем корень из операнда
 }

 else if (clickedOperator == tr("1/x")) { //следующая функция занимается делением единицы на операнд
  if (operand == 0.0) {
   abortOperation();
   return;
  }
  result = 1.0 / operand;
 }
 ui->lcd1->setText(QString::number(result)); //Выводим result на экран
 waitingForOperand = true;                   //ждем следующего операнда
}
//Заканчиваем с унарными операторами

//Здесь мы как обычно слушаем sender и считываем что написано на кнопках из GUI
void madCalc::Operator(){
    QPushButton *button = (QPushButton *)sender();
    clickedOperator = button->text(); //А затем записываем полученные значения в в clickedOperator
    OperatorSender(); //И вызываем функцию OperatorSender
}

//Данная функция определяет что делать с операторами сложения/вычитания и умножения/деления
void madCalc::OperatorSender(){
    if (clickedOperator == "+"||"-")      {additiveOperator();} //Для операторов сложения вызываем это
    else if (clickedOperator == "*"||"/") {multiplicativeOperator();} //Для мультипликативных это
    QString string = QString::number(num);
    ui->lcd2->setText(string.append(clickedOperator)); //Приращиваем к содержимому LCD2 кликнутый оператор
    LCD2=string;
}

//Тут происходит неведомая фигня... Попробуй, разберись )
void madCalc::additiveOperator() {
    ui->lcd3->setText("");

 double operand = ui->lcd1->text().toDouble();              //Объявляем operand, содержащий циферьки с lcd1
 if (!pendingMultiplicativeOperator.isEmpty()) {            //Запускаем проверку: если переменная оператора не пуста, то
  if (!calculate(operand, pendingMultiplicativeOperator)) { //запускаем проверку: если calculate не содержит operand и оператор, то
   abortOperation(); //Функция аборта
   return;
  }
  ui->lcd1->setText(QString::number(num)); //записываем на экран содержимое переменной num
  operand = num;                           //приравниваем operand к num
  num = 0.0;                               //А теперь num к нулю
  pendingMultiplicativeOperator.clear();   //И очищаем переменную, содержащую знак текущего оператора
 }
 if (!pendingAdditiveOperator.isEmpty()) { //Если эта переменная не пуста, то аборт
  if (!calculate(operand, pendingAdditiveOperator)) {
   abortOperation();
   return;
  }
  ui->lcd1->setText(QString::number(num));      //выводим num на экран
 }
 else {
  num = operand;                                //Записываем operand в num
 }
 pendingAdditiveOperator = clickedOperator;
 waitingForOperand = true;                      //Ждем следующего операнда
}
//Закончили с  операторами сложения

//Мультипликативные операторы. Функция работает примерно так же как так предыдущая
void madCalc::multiplicativeOperator() {
    ui->lcd3->setText("");

 double operand = ui->lcd1->text().toDouble(); //Берем операнд с экрана, переводим его в переменную типа double
 if (!pendingMultiplicativeOperator.isEmpty()) {
  if (!calculate(operand, pendingMultiplicativeOperator)) {
   abortOperation();
   return;
  }
  ui->lcd1->setText(QString::number(num));
 }
 else {
  num = operand;
 }
 pendingMultiplicativeOperator = clickedOperator;
 waitingForOperand = true;
}

//Функция знака =
void madCalc::equalClicked() {
 double operand = ui->lcd1->text().toDouble();
 if (!pendingMultiplicativeOperator.isEmpty()) {
  if (!calculate(operand, pendingMultiplicativeOperator)) {
   abortOperation();
   return;
  }
  operand = num;
  num = 0.0;
  pendingMultiplicativeOperator.clear();
 }
 if (!pendingAdditiveOperator.isEmpty()) {
  if (!calculate(operand, pendingAdditiveOperator)) {
   abortOperation();
   return;
  }
  pendingAdditiveOperator.clear();
 }
 else {
  num = operand;
 }
 ui->lcd1->setText(QString::number(num));
 num= 0.0;

 //Конструкция для LCD2
 QString string = QString::number(operand);  //string получает значение operand
 ui->lcd2->setText(LCD2.append(string)+"=");
 LCD2="";
 waitingForOperand = true;
}

//функция вычислений
bool madCalc::calculate(double rightOperand, const QString &pendingOperator) {
 if (pendingOperator == tr("+")) num += rightOperand;
 else if (pendingOperator == tr("-")) num -= rightOperand;
 else if (pendingOperator == tr("*")) num *= rightOperand;
 else if (pendingOperator == tr("/")) {
  if (rightOperand == 0.0) return false;
  num /= rightOperand;
   }
  return true;
  ui->lcd3->setText("");
}

void madCalc::CE()
{
    ui->lcd3->setText("");
    if (waitingForOperand) return;
    ui->lcd1->setText("0");
    waitingForOperand = true;

}

void madCalc::ClearAll()
{
    num = 0.0; //Приравниваем предыдущий операнд к нулю
    pendingAdditiveOperator.clear(); //Очищаем переменную, отвечающую за хранения знака оператора сложения/вычитания
    pendingMultiplicativeOperator.clear(); //Очищаем переменную, отвечающую за хранения знака оператора умножения/деления
    ui->lcd1->setText("0"); //Устанавливаем 0 на наш экран
    ui->lcd2->setText("");
    ui->lcd3->setText("");
    waitingForOperand = true; //Ждем следующего операнда
}

//Вызов формы с информацией о программе
void madCalc::About()
{
    formAbout->show();
}
